from sphinx.domains import Domain

from . import directive, role


class SpellingDomain(Domain):
    name = "spelling"
    label = "Spelling Checker"
    directives = {
        "word-list": directive.SpellingDirective,
    }
    roles = {"word": role.spelling_word, "ignore": role.spelling_ignore}

    def get_objects(self):
        return []

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        """resolve cross reference"""
        return None

    def resolve_any_xref(self, env, fromdocname, builder, target, node, contnode):
        return []

    def merge_domaindata(self, docnames, otherdata):
        pass



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Tests for SpellingBuilder"""

import contextlib
import io
import os
import sys
import textwrap

import pytest
from sphinx.application import Sphinx

from tests import helpers  # isort:skip


def _make_sphinx_project(tmpdir):
    srcdir = tmpdir.mkdir("src")
    outdir = tmpdir.mkdir("out")
    add_file(
        srcdir,
        "conf.py",
        """
    extensions = [ 'sphinxcontrib.spelling' ]
    """,
    )
    return (srcdir, outdir)


@pytest.fixture
def sphinx_project(tmpdir):
    yield _make_sphinx_project(tmpdir)


@contextlib.contextmanager
def working_dir(targetdir):
    "Temporarily change the working directory of the process."
    before = os.getcwd()
    os.chdir(targetdir)
    try:
        yield os.getcwd()
    finally:
        os.chdir(before)


@contextlib.contextmanager
def import_path(new_path):
    "Temporarily change sys.path for imports."
    before = sys.path
    try:
        sys.path = new_path
        yield
    finally:
        sys.path = before


def add_file(thedir, filename, content):
    with open(thedir.join(filename), "w") as f:
        f.write(textwrap.dedent(content))


def get_sphinx_app(srcdir, outdir, docname, builder="spelling"):
    stdout = io.StringIO()
    stderr = io.StringIO()
    app = Sphinx(
        srcdir,
        srcdir,
        outdir,
        outdir,
        builder,
        status=stdout,
        warning=stderr,
        freshenv=True,
    )
    return (stdout, stderr, app)


def get_sphinx_output(srcdir, outdir, docname, builder="spelling"):
    stdout, stderr, app = get_sphinx_app(srcdir, outdir, docname, builder)
    app.build()
    path = os.path.join(outdir, f"{docname}.spelling")
    try:
        with open(path, "r") as f:
            output_text = f.read()
    except FileNotFoundError:
        output_text = None
    return (stdout, stderr, output_text)


def test_setup(sphinx_project):
    srcdir, outdir = sphinx_project
    stdout = io.StringIO()
    stderr = io.StringIO()
    # If the spelling builder is not properly initialized,
    # trying to use it with the Sphinx app class will
    # generate an exception.
    Sphinx(
        str(srcdir),
        str(srcdir),
        str(outdir),
        str(outdir),
        "spelling",
        status=stdout,
        warning=stderr,
        freshenv=True,
    )


def test_title(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "contents.rst",
        """
    Welcome to Speeling Checker documentation!
    ==========================================
    """,
    )
    stdout, stderr, output_text = get_sphinx_output(srcdir, outdir, "contents")
    assert "(Speeling)" in output_text


def test_body(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "contents.rst",
        """
    Welcome to Spelling Checker documentation!
    ==========================================

    There are several mispelled words in this txt.
    """,
    )
    stdout, stderr, output_text = get_sphinx_output(srcdir, outdir, "contents")
    assert "(mispelled)" in output_text
    assert "(txt)" in output_text


def test_ignore_literals(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "contents.rst",
        """
    Welcome to Spelling Checker documentation!
    ==========================================

    There are several misspelled words in this text.

    ::

        Literal blocks are ignoreed.

    Inline ``litterals`` are ignored, too.

    """,
    )
    stdout, stderr, output_text = get_sphinx_output(srcdir, outdir, "contents")
    # The 'contents.spelling' output file should not have been
    # created, because the errors are ignored.
    assert output_text is None


def test_several_word_lists(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "conf.py",
        """
    extensions = ['sphinxcontrib.spelling']
    spelling_word_list_filename=['test_wordlist.txt','test_wordlist2.txt']
    """,
    )

    add_file(
        srcdir,
        "contents.rst",
        """
    Welcome to Spelling Checker documentation!
    ==========================================

    There are several mispelled words in tihs txt.
    """,
    )

    add_file(
        srcdir,
        "test_wordlist.txt",
        """
    txt
    """,
    )

    add_file(
        srcdir,
        "test_wordlist2.txt",
        """
    mispelled
    """,
    )
    stdout, stderr, output_text = get_sphinx_output(srcdir, outdir, "contents")
    # Both of these should be fine now
    assert "(mispelled)" not in output_text
    assert "(txt)" not in output_text
    # But not this one
    assert "(tihs)" in output_text


def _wordlist_sphinx_project(tmpdir, conf_contents):
    srcdir, outdir = _make_sphinx_project(tmpdir)
    add_file(srcdir, "conf.py", conf_contents)
    add_file(
        srcdir,
        "test_wordlist.txt",
        """
    txt
    """,
    )
    add_file(
        srcdir,
        "test_wordlist2.txt",
        """
    mispelled
    """,
    )
    stdout, stderr, app = get_sphinx_app(srcdir, outdir, "contents")
    return (srcdir, outdir, stdout, stderr, app)


def test_word_list_default(tmpdir):
    srcdir, outdir, stdout, stderr, app = _wordlist_sphinx_project(
        tmpdir,
        """
        extensions = ['sphinxcontrib.spelling']
        """,
    )
    results = app.builder.get_configured_wordlist_filenames()
    assert len(results) == 1
    assert os.path.basename(results[0]) == "spelling_wordlist.txt"


def test_one_word_list_str(tmpdir):
    srcdir, outdir, stdout, stderr, app = _wordlist_sphinx_project(
        tmpdir,
        """
        extensions = ['sphinxcontrib.spelling']
        spelling_word_list_filename='test_wordlist.txt'
        """,
    )
    results = app.builder.get_configured_wordlist_filenames()
    assert len(results) == 1
    assert os.path.basename(results[0]) == "test_wordlist.txt"


def test_multiple_word_list_str(tmpdir):
    # We don't expect anyone to set up their conf.py this way but it
    # simulates passing the configuration option from the command line
    # using -D.
    srcdir, outdir, stdout, stderr, app = _wordlist_sphinx_project(
        tmpdir,
        """
        extensions = ['sphinxcontrib.spelling']
        spelling_word_list_filename='test_wordlist.txt,test_wordlist2.txt'
        """,
    )
    results = app.builder.get_configured_wordlist_filenames()
    assert len(results) == 2
    assert os.path.basename(results[0]) == "test_wordlist.txt"
    assert os.path.basename(results[1]) == "test_wordlist2.txt"


def test_multiple_word_list_list(tmpdir):
    srcdir, outdir, stdout, stderr, app = _wordlist_sphinx_project(
        tmpdir,
        """
        extensions = ['sphinxcontrib.spelling']
        spelling_word_list_filename=['test_wordlist.txt', 'test_wordlist2.txt']
        """,
    )
    results = app.builder.get_configured_wordlist_filenames()
    assert len(results) == 2
    assert os.path.basename(results[0]) == "test_wordlist.txt"
    assert os.path.basename(results[1]) == "test_wordlist2.txt"


def test_ignore_file(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "conf.py",
        """
    extensions = ['sphinxcontrib.spelling']
    spelling_exclude_patterns=['con*']
    """,
    )

    add_file(
        srcdir,
        "contents.rst",
        """
    Welcome to Speeling Checker documentation!
    ==========================================
    """,
    )

    stdout, stderr, output_text = get_sphinx_output(srcdir, outdir, "contents")
    # The 'contents.spelling' output file should not have been
    # created, because the file is ignored.
    assert output_text is None


@helpers.require_git_repo
def test_docstrings(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "conf.py",
        """
    extensions = ['sphinxcontrib.spelling', 'sphinx.ext.autodoc']
    """,
    )

    add_file(
        srcdir / "..",
        "the_source.py",
        '''
    #!/usr/bin/env python3

    def public_function(arg_name):
        """Does something useful.

        :param arg_name: Pass a vaule
        """
        return 1
    ''',
    )

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    .. automodule:: the_source
       :members:

    """,
    )

    with working_dir(srcdir / ".."):
        with import_path(["."] + sys.path):
            stdout, stderr, output_text = get_sphinx_output(
                srcdir,
                outdir,
                "contents",
            )

    expected = "src/contents.rst:3: (vaule)  Pass a vaule\n"
    assert expected in output_text


def test_get_suggestions_to_show_all(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "conf.py",
        """
    extensions = ['sphinxcontrib.spelling']
    spelling_show_suggestions = True
    spelling_suggestion_limit = 0
    """,
    )
    stdout, stderr, app = get_sphinx_app(srcdir, outdir, "contents")
    results = app.builder.get_suggestions_to_show(["a", "b", "c"])
    assert len(results) == 3


def test_get_suggestions_to_show_limit(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "conf.py",
        """
    extensions = ['sphinxcontrib.spelling']
    spelling_show_suggestions = True
    spelling_suggestion_limit = 1
    """,
    )
    stdout, stderr, app = get_sphinx_app(srcdir, outdir, "contents")
    results = app.builder.get_suggestions_to_show(["a", "b", "c"])
    assert len(results) == 1


def test_get_suggestions_to_show_disabled(sphinx_project):
    srcdir, outdir = sphinx_project
    add_file(
        srcdir,
        "conf.py",
        """
    extensions = ['sphinxcontrib.spelling']
    spelling_show_suggestions = False
    spelling_suggestion_limit = 0
    """,
    )
    stdout, stderr, app = get_sphinx_app(srcdir, outdir, "contents")
    results = app.builder.get_suggestions_to_show(["a", "b", "c"])
    assert len(results) == 0


def test_captions(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    .. figure:: blah.gif

       Teh caption

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert "(Teh)" in output_text


def test_legacy_directive(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    .. spelling::

       teh

    teh is OK

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert output_text is None


def test_domain_directive(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    .. spelling:word-list::

       teh

    teh is OK

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert output_text is None


def test_domain_role(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    :spelling:word:`teh` is OK

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert output_text is None


def test_domain_role_multiple_words(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    :spelling:word:`teh is KO`

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert output_text is None


def test_domain_role_output(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    :spelling:word:`teh` is OK

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
        "text",
    )

    path = os.path.join(outdir, "contents.txt")
    try:
        with open(path, "r") as f:
            output_text = f.read()
    except FileNotFoundError:
        output_text = None

    assert output_text == "The Module\n**********\n\nteh is OK\n"


def test_domain_ignore(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    :spelling:ignore:`baddddd` is OK

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert output_text is None


def test_domain_ignore_multiple_words(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    :spelling:ignore:`baddddd` is OK here.

    But, baddddd is not OK here.
    Nor, here baddddd.

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert "(baddddd)" in output_text
    assert output_text.count("\n") == 2  # Only expect 2 errors, not 3.


def test_domain_ignore_output(sphinx_project):
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    :spelling:ignore:`teh` is OK

    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
        "text",
    )

    path = os.path.join(outdir, "contents.txt")
    try:
        with open(path, "r") as f:
            output_text = f.read()
    except FileNotFoundError:
        output_text = None

    assert output_text == "The Module\n**********\n\nteh is OK\n"


def test_only_directive(sphinx_project):
    # How to skip checking nested blocks of content
    # https://github.com/sphinx-contrib/spelling/issues/204
    srcdir, outdir = sphinx_project

    add_file(
        srcdir,
        "contents.rst",
        """
    The Module
    ==========

    .. only:: html

       teh is ok

    whaat is not ok
    """,
    )

    stdout, stderr, output_text = get_sphinx_output(
        srcdir,
        outdir,
        "contents",
    )
    assert "(whaat)" in output_text
    assert "(teh)" not in output_text


def test_nodes_with_none_source(sphinx_project):
    # Reproduces https://github.com/sphinx-contrib/spelling/issues/234
    # Tests handling of nodes where get_source_line returns None for source
    srcdir, outdir = sphinx_project

    # Create an extension that patches docutils.utils.get_source_line to return None
    # This directly simulates the bug condition
    add_file(
        srcdir,
        "mock_none_source.py",
        """
import docutils.utils


# Store the original function
_original_get_source_line = docutils.utils.get_source_line


def patched_get_source_line(node):
    '''Return None for source to simulate issue #234'''
    # Call original to get line number
    source, lineno = _original_get_source_line(node)
    # But return None for source to trigger the bug
    return (None, lineno)


def setup(app):
    # Patch the function when building
    def on_build_finished(app, exception):
        # Restore original after build
        docutils.utils.get_source_line = _original_get_source_line

    # Patch before doctree-read
    def on_doctree_read(app, doctree):
        docutils.utils.get_source_line = patched_get_source_line

    app.connect('doctree-read', on_doctree_read, priority=1)
    app.connect('build-finished', on_build_finished)

    return {'version': '0.1'}
    """,
    )

    add_file(
        srcdir,
        "conf.py",
        f"""
import sys
sys.path.insert(0, r'{srcdir}')
extensions = ['sphinxcontrib.spelling', 'mock_none_source']
    """,
    )

    add_file(
        srcdir,
        "contents.rst",
        """
    Test Document
    =============

    This text has a mispeling that should be caught.
    """,
    )

    # Without the fix, this will crash with:
    # TypeError: expected str, bytes or os.PathLike object, not NoneType
    # With the fix, it should handle None gracefully and produce output
    stdout, stderr, output_text = get_sphinx_output(srcdir, outdir, "contents")

    # The spelling check should still work and find the misspelling
    assert output_text is not None
    assert "(mispeling)" in output_text



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Spelling checker extension for Sphinx."""

import collections
import importlib
import os
import tempfile

import docutils.nodes
import docutils.utils
from sphinx.builders import Builder
from sphinx.util import logging, osutil
from sphinx.util.console import red
from sphinx.util.matching import Matcher
from sphinx.util.osutil import ensuredir

try:
    from enchant.tokenize import EmailFilter, WikiWordFilter
except ImportError as imp_exc:
    enchant_import_error = imp_exc
else:
    enchant_import_error = None

from . import checker, filters

logger = logging.getLogger(__name__)

# TODO - Words with multiple uppercase letters treated as classes and ignored


class SpellingBuilder(Builder):
    """
    Spell checks a document
    """

    name = "spelling"

    def init(self):
        if enchant_import_error is not None:
            raise RuntimeError(
                "Cannot initialize spelling builder without PyEnchant installed"
            ) from enchant_import_error
        self.misspelling_count = 0

        self.env.settings["smart_quotes"] = False
        # Initialize the per-document filters
        if not hasattr(self.env, "spelling_document_words"):
            self.env.spelling_document_words = collections.defaultdict(list)

        # Initialize the global filters
        f = [
            filters.ContractionFilter,
            EmailFilter,
        ]
        if self.config.spelling_ignore_wiki_words:
            logger.info("Ignoring wiki words")
            f.append(WikiWordFilter)
        if self.config.spelling_ignore_acronyms:
            logger.info("Ignoring acronyms")
            f.append(filters.AcronymFilter)
        if self.config.spelling_ignore_pypi_package_names:
            logger.info("Adding package names from PyPI to local dictionary…")
            f.append(filters.PyPIFilterFactory())
        if self.config.spelling_ignore_python_builtins:
            logger.info("Ignoring Python builtins")
            f.append(filters.PythonBuiltinsFilter)
        if self.config.spelling_ignore_importable_modules:
            logger.info("Ignoring importable module names")
            f.append(filters.ImportableModuleFilter)
        if self.config.spelling_ignore_contributor_names:
            logger.info("Ignoring contributor names")
            f.append(filters.ContributorFilter)
        f.extend(self._load_filter_classes(self.config.spelling_filters))

        if not os.path.isdir(self.outdir):
            os.mkdir(self.outdir)

        word_list = self.get_wordlist_filename()
        logger.info("Looking for custom word list in %s", word_list)

        self.checker = checker.SpellingChecker(
            lang=self.config.spelling_lang,
            tokenizer_lang=self.config.tokenizer_lang,
            suggest=self.config.spelling_show_suggestions,
            word_list_filename=word_list,
            filters=f,
            context_line=self.config.spelling_show_whole_line,
        )

    def _load_filter_classes(self, filters):
        # Filters may be expressed in the configuration file using
        # names, so look through them and import the referenced class
        # and use that in the checker.
        for filter_ in filters:
            if not isinstance(filter_, str):
                yield filter_
                continue
            module_name, _, class_name = filter_.rpartition(".")
            mod = importlib.import_module(module_name)
            yield getattr(mod, class_name)

    def get_configured_wordlist_filenames(self):
        "Returns the configured wordlist filenames."
        word_list = self.config.spelling_word_list_filename
        if word_list is None:
            word_list = ["spelling_wordlist.txt"]

        if isinstance(word_list, str):
            # Wordlist is a string. Split on comma in case it came
            # from the command line, via -D, and has multiple values.
            word_list = word_list.split(",")

        return [os.path.join(self.srcdir, p) for p in word_list]

    def get_wordlist_filename(self):
        "Returns the filename of the wordlist to use when checking content."
        filenames = self.get_configured_wordlist_filenames()
        if len(filenames) == 1:
            return filenames[0]
        # In case the user has multiple word lists, we combine them
        # into one large list that we pass on to the checker.
        return self._build_combined_wordlist()

    def _build_combined_wordlist(self):
        # If we have a list, the combined list is the first list plus all words
        # from the other lists. Otherwise, word_list is assumed to just be a
        # string.
        temp_dir = tempfile.mkdtemp()
        combined_word_list = os.path.join(temp_dir, "spelling_wordlist.txt")

        with open(combined_word_list, "w", encoding="UTF-8") as outfile:
            for word_file in self.get_configured_wordlist_filenames():
                # Paths are relative
                long_word_file = os.path.join(self.srcdir, word_file)
                logger.info("Adding contents of %s to custom word list", long_word_file)
                with open(long_word_file, encoding="UTF-8") as infile:
                    infile_contents = infile.readlines()
                outfile.writelines(infile_contents)

                # Check for newline, and add one if not present
                if infile_contents and not infile_contents[-1].endswith("\n"):
                    outfile.write("\n")

        return combined_word_list

    def get_outdated_docs(self):
        return "all documents"

    def prepare_writing(self, docnames):
        return

    def get_target_uri(self, docname, typ=None):
        return ""

    def get_suggestions_to_show(self, suggestions):
        if not self.config.spelling_show_suggestions or not suggestions:
            return []
        to_show = suggestions
        try:
            n_to_show = int(self.config.spelling_suggestion_limit)
        except ValueError:
            n_to_show = 0
        if n_to_show > 0:
            to_show = suggestions[:n_to_show]
        return to_show

    def format_suggestions(self, suggestions):
        to_show = self.get_suggestions_to_show(suggestions)
        if not to_show:
            return ""
        return "[" + ", ".join('"%s"' % s for s in to_show) + "]"

    TEXT_NODES = {
        "block_quote",
        "caption",
        "paragraph",
        "list_item",
        "term",
        "definition_list_item",
        "title",
    }

    def write_doc(self, docname, doctree):
        """write the document"""
        lines = list(self._find_misspellings(docname, doctree))
        self.misspelling_count += len(lines)
        if lines:
            output_filename = os.path.join(self.outdir, f"{docname}.spelling")
            logger.info("Writing %s", output_filename)
            ensuredir(os.path.dirname(output_filename))
            with open(output_filename, "w", encoding="UTF-8") as output:
                output.writelines(lines)

    def _find_misspellings(self, docname, doctree):
        excluded = Matcher(self.config.spelling_exclude_patterns)
        if excluded(self.env.doc2path(docname, None)):
            return
        # Build the document-specific word filter based on any good
        # words listed in spelling directives. If we have no such
        # words, we want to push an empty list of filters so that we
        # can always safely pop the filter stack when we are done with
        # this document.
        doc_filters = []
        good_words = self.env.spelling_document_words.get(docname)
        if good_words:
            logger.debug("Extending local dictionary for %s", docname)
            doc_filters.append(filters.IgnoreWordsFilterFactory(good_words))
        self.checker.push_filters(doc_filters)

        # Set up a filter for the types of nodes to ignore during
        # traversal.
        def filter(n):
            if n.tagname != "#text":
                return False
            if n.parent and n.parent.tagname not in self.TEXT_NODES:
                return False
            # Nodes marked by the spelling:ignore role
            if hasattr(n, "spellingIgnore"):
                return False
            return True

        for node in doctree.findall(filter):
            # Get the location of the text being checked so we can
            # report it in the output file. Nodes from text that
            # comes in via an 'include' directive does not include
            # the full path, so convert all to relative path
            # for consistency.
            source, node_lineno = docutils.utils.get_source_line(node)
            if source is not None:
                source = osutil.relpath(source)
            else:
                # Some nodes (e.g., programmatically generated) may not have
                # source information. Use a placeholder.
                source = "<unknown>"

            # Check the text of the node.
            misspellings = self.checker.check(node.astext())
            for word, suggestions, context_line, line_offset in misspellings:
                # Avoid TypeError on nodes lacking a line number
                # This happens for some node originating from docstrings
                lineno = node_lineno
                if lineno is not None:
                    lineno += line_offset

                msg_parts = [
                    f"{source}:{lineno}: ",
                    "Spell check",
                    red(word),
                ]
                if self.format_suggestions(suggestions) != "":
                    msg_parts.append(self.format_suggestions(suggestions))
                msg_parts.append(context_line)
                msg = ": ".join(msg_parts) + "."
                if self.config.spelling_warning:
                    logger.warning(msg)
                elif self.config.spelling_verbose:
                    logger.info(msg)
                yield "%s:%s: (%s) %s %s\n" % (
                    source,
                    lineno,
                    word,
                    self.format_suggestions(suggestions),
                    context_line,
                )

        self.checker.pop_filters()
        return

    def finish(self):
        if self.misspelling_count:
            logger.warning("Found %d misspelled words", self.misspelling_count)



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Spelling checker extension for Sphinx."""

try:
    import enchant
    from enchant.tokenize import get_tokenizer
except ImportError as imp_exc:
    enchant_import_error = imp_exc
else:
    enchant_import_error = None


class SpellingChecker:
    """Checks the spelling of blocks of text.

    Uses options defined in the sphinx configuration file to control
    the checking and filtering behavior.
    """

    def __init__(
        self,
        lang,
        suggest,
        word_list_filename,
        tokenizer_lang="en_US",
        filters=None,
        context_line=False,
    ):
        if enchant_import_error is not None:
            raise RuntimeError(
                "Cannot instantiate SpellingChecker without PyEnchant installed",
            ) from enchant_import_error
        if filters is None:
            filters = []
        self.dictionary = enchant.DictWithPWL(lang, word_list_filename)
        self.tokenizer = get_tokenizer(tokenizer_lang, filters=filters)
        self.original_tokenizer = self.tokenizer
        self.suggest = suggest
        self.context_line = context_line

    def push_filters(self, new_filters):
        """Add a filter to the tokenizer chain."""
        t = self.tokenizer
        for f in new_filters:
            t = f(t)
        self.tokenizer = t

    def pop_filters(self):
        """Remove the filters pushed during the last call to push_filters()."""
        self.tokenizer = self.original_tokenizer

    def check(self, text):
        """Yields bad words and suggested alternate spellings."""
        for word, pos in self.tokenizer(text):
            correct = self.dictionary.check(word)
            if correct:
                continue

            suggestions = self.dictionary.suggest(word) if self.suggest else []
            line = line_of_index(text, pos) if self.context_line else ""
            line_offset = text.count("\n", 0, pos)

            yield word, suggestions, line, line_offset


def line_of_index(text, index):
    try:
        line_start = text.rindex("\n", 0, index) + 1
    except ValueError:
        line_start = 0
    try:
        line_end = text.index("\n", index)
    except ValueError:
        line_end = len(text)

    return text[line_start:line_end]



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Spelling checker extension for Sphinx."""

# TODO - Words with multiple uppercase letters treated as classes and ignored

import builtins
import importlib
import subprocess
import sys

import requests
from enchant.tokenize import Filter, get_tokenizer, tokenize, unit_tokenize
from sphinx.util import logging

logger = logging.getLogger(__name__)


class AcronymFilter(Filter):
    """If a word looks like an acronym (all upper case letters),
    ignore it.
    """

    def _skip(self, word):
        return (
            word.isupper()  # all caps
            or
            # pluralized acronym ("URLs")
            (word[-1].lower() == "s" and word[:-1].isupper())
        )


class list_tokenize(tokenize):
    def __init__(self, words):
        super().__init__("")
        self._words = words

    def next(self):
        if not self._words:
            raise StopIteration()
        word = self._words.pop(0)
        return (word, 0)


class ContractionFilter(Filter):
    """Strip common contractions from words."""

    splits = {
        "aren't": ["are", "not"],
        "can't": ["can", "not"],
        "could've": ["could", "have"],
        "couldn't": ["could", "not"],
        "didn't": ["did", "not"],
        "doesn't": ["does", "not"],
        "don't": ["do", "not"],
        "hadn't": ["had", "not"],
        "hasn't": ["has", "not"],
        "haven't": ["have", "not"],
        "he'd": ["he", "would"],
        "he'll": ["he", "will"],
        "he's": ["he", "is"],
        "how'd": ["how", "would"],
        "how'll": ["how", "will"],
        "how's": ["how", "is"],
        "i'd": ["I", "would"],
        "i'll": ["I", "will"],
        "i'm": ["I", "am"],
        "i've": ["I", "have"],
        "isn't": ["is", "not"],
        "it'd": ["it", "would"],
        "it'll": ["it", "will"],
        "it's": ["it", "is"],
        "ma'am": ["madam"],
        "might've": ["might", "have"],
        "mightn't": ["might", "not"],
        "must've": ["must", "have"],
        "mustn't": ["must", "not"],
        "o'": ["of"],
        "o'clock": ["of", "the", "clock"],
        "she'd": ["she", "would"],
        "she'll": ["she", "will"],
        "she's": ["she", "is"],
        "should've": ["should", "have"],
        "shouldn't": ["should", "not"],
        "that'd": ["that", "would"],
        "that'll": ["that", "will"],
        "that's": ["that", "is"],
        "they'd": ["they", "would"],
        "they'll": ["they", "will"],
        "they're": ["they", "are"],
        "they've": ["they", "have"],
        "wasn't": ["was", "not"],
        "we'd": ["we", "would"],
        "we'll": ["we", "will"],
        "we're": ["we", "are"],
        "we've": ["we", "have"],
        "weren't": ["were", "not"],
        "what'd": ["what", "would"],
        "what'll": ["what", "will"],
        "what're": ["what", "are"],
        "what's": ["what", "is"],
        "when'd": ["when", "would"],
        "when'll": ["when", "will"],
        "when's": ["when", "is"],
        "where'd": ["where", "would"],
        "where'll": ["where", "will"],
        "where's": ["where", "is"],
        "who'd": ["who", "would"],
        "who'll": ["who", "will"],
        "who's": ["who", "is"],
        "why'd": ["why", "would"],
        "why'll": ["why", "will"],
        "why's": ["why", "is"],
        "won't": ["will", "not"],
        "would've": ["would", "have"],
        "wouldn't": ["would", "not"],
        "you'd": ["you", "would"],
        "you'll": ["you", "will"],
        "you're": ["you", "are"],
        "you've": ["you", "have"],
    }

    def _split(self, word):
        # Fixed responses
        if word.lower() in self.splits:
            return list_tokenize(self.splits[word.lower()])

        # Possessive
        if word.lower().endswith("'s"):
            return unit_tokenize(word[:-2])

        # * not
        if word.lower().endswith("n't"):
            return unit_tokenize(word[:-3])

        return unit_tokenize(word)


class IgnoreWordsFilter(Filter):
    """Given a set of words, ignore them all."""

    def __init__(self, tokenizer, word_set):
        self.word_set = set(word_set)
        super().__init__(tokenizer)

    def _skip(self, word):
        return word in self.word_set


class IgnoreWordsFilterFactory:
    def __init__(self, words):
        self.words = words

    def __call__(self, tokenizer):
        return IgnoreWordsFilter(tokenizer, self.words)


class PyPIFilterFactory(IgnoreWordsFilterFactory):
    """Build an IgnoreWordsFilter for all of the names of packages on PyPI."""

    def __init__(self):
        r = requests.get(
            "https://pypi.org/simple/",
            headers={
                "user-agent": "sphinxcontrib.spelling",
                "accept": "application/vnd.pypi.simple.v1+json",
            },
        )
        names = [i["name"] for i in r.json()["projects"]]
        logger.debug("retrieved %d project names from pypi.org", len(names))
        super().__init__(names)


class PythonBuiltinsFilter(Filter):
    """Ignore names of built-in Python symbols."""

    def _skip(self, word):
        return hasattr(builtins, word)


class ImportableModuleFilter(Filter):
    """Ignore names of modules that we could import."""

    def __init__(self, tokenizer):
        super().__init__(tokenizer)
        self.found_modules = set(sys.builtin_module_names)
        self.sought_modules = self.found_modules.copy()
        # By adding __main__ to the list of sought modules but not
        # found modules we ensure that it is never recognized as a
        # valid module, which is consistent with the behavior before
        # version 7.3.1.  See
        # https://github.com/sphinx-contrib/spelling/issues/141
        self.sought_modules.add("__main__")

    def _skip(self, word):
        # If the word looks like a python module filename, strip the
        # extension to avoid the side-effect of actually importing the
        # module. This prevents, for example, 'setup.py' triggering an
        # import of the setup module during a doc build, which makes
        # it look like Sphinx is complaining about a commandline
        # argument. See
        # https://github.com/sphinx-contrib/spelling/issues/142
        if word.endswith(".py"):
            logger.debug(
                "removing .py extension from %r before searching for module", word
            )
            word = word[:-3]

        valid_module_name = all(n.isidentifier() for n in word.split("."))
        if not valid_module_name:
            return False

        if word not in self.sought_modules:
            self.sought_modules.add(word)
            try:
                mod = importlib.util.find_spec(word)
            except BaseException as err:
                # This could be an ImportError, SystemExit, some more detailed
                # error out of distutils, or something else triggered
                # by failing to be able to import a parent package to
                # use the metadata to search for a subpackage.
                logger.debug("find_spec(%r) failed, invalid module name: %s", word, err)
            else:
                if mod is not None:
                    self.found_modules.add(word)

        return word in self.found_modules


class ContributorFilter(IgnoreWordsFilter):
    """Accept information about contributors as spelled correctly.

    Look in the git history for authors and committers and accept
    tokens that are in the set.
    """

    _pretty_format = "%(trailers:key=Co-Authored-By,separator=%x0A)%x0A%an%x0A%cn"

    def __init__(self, tokenizer):
        contributors = self._get_contributors()
        super().__init__(tokenizer, contributors)

    def _get_contributors(self):
        logger.info("Scanning contributors")
        cmd = [
            "git",
            "log",
            "--quiet",
            "--no-color",
            f"--pretty=format:{self._pretty_format}",
        ]

        try:
            p = subprocess.run(cmd, check=True, stdout=subprocess.PIPE)
        except (subprocess.CalledProcessError, FileNotFoundError) as err:
            logger.warning("Called: %s", " ".join(cmd))
            logger.warning("Failed to scan contributors: %s", err)
            return set()
        output = p.stdout.decode("utf-8")
        tokenizer = get_tokenizer("en_US", filters=[])
        return {word for word, pos in tokenizer(output)}



from docutils import nodes

from . import directive


def spelling_word(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """Let the user indicate that inline text is spelled correctly."""
    env = inliner.document.settings.env
    docname = env.docname
    good_words = text.split()
    directive.add_good_words_to_document(env, docname, good_words)
    node = nodes.Text(text)
    return [node], []


def spelling_ignore(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """Let the user indicate that inline text is to not be spellchecked."""
    node = nodes.Text(text)
    setattr(node, "spellingIgnore", True)
    return [node], []



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Tests for filters."""

import contextlib
import logging
import os
import sys

import pytest
from enchant.tokenize import get_tokenizer

from sphinxcontrib.spelling import filters  # isort:skip
from tests import helpers  # isort:skip

# Replace the sphinx logger with a normal one so pytest can collect
# the output.
filters.logger = logging.getLogger("test.filters")


def test_builtin_unicode():
    f = filters.PythonBuiltinsFilter(None)
    assert not f._skip("passé")


def test_builtin_regular():
    f = filters.PythonBuiltinsFilter(None)
    assert f._skip("print")


def test_acronym():
    text = "a front-end for DBM-style databases"
    t = get_tokenizer("en_US", [])
    f = filters.AcronymFilter(t)
    words = [w[0] for w in f(text)]
    assert "DBM" not in words, "Failed to filter out acronym"


def test_acronym_unicode():
    text = "a front-end for DBM-style databases"
    t = get_tokenizer("en_US", [])
    f = filters.AcronymFilter(t)
    words = [w[0] for w in f(text)]
    assert "DBM" not in words, "Failed to filter out acronym"


@helpers.require_git_repo
@pytest.mark.parametrize(
    "name",
    [
        "Alex",
        "Atlakson",
        "Avram",
        "Baumgold",
        "Berman",
        "Daniele",
        "Doug",
        "Finucane",
        "Gaynor",
        "Gonsiorowski",
        "Hong",
        "Hong",
        "Huon",
        "Kampik",
        "Kolosov",
        "Lubkin",
        "Marti",
        "Minhee",
        "Olausson",
        "Raggam",
        "Raudsepp",
        "sdelliot",
        "Sergey",
        "Sevilla",
        "Timotheus",
        "Tobias",
        "Tricoli",
    ],
)
def test_contributors(name):
    f = filters.ContributorFilter(None)
    assert f._skip(name)


@pytest.mark.parametrize(
    "word,expected",
    [
        ("os", True),
        ("os.name", False),
        ("__main__", False),
        ("don't", False),
    ],
)
def test_importable_module_skip(word, expected):
    f = filters.ImportableModuleFilter(None)
    assert f._skip(word) is expected


@contextlib.contextmanager
def import_path(new_path):
    "Temporarily change sys.path for imports."
    before = sys.path
    try:
        sys.path = new_path
        yield
    finally:
        sys.path = before


def test_importable_module_with_side_effets(tmpdir):
    logging.debug("tmpdir %r", tmpdir)
    logging.debug("cwd %r", os.getcwd())

    parentdir = tmpdir.join("parent")
    parentdir.mkdir()

    parentdir.join("__init__.py").write('raise SystemExit("exit as side-effect")\n')
    parentdir.join("child.py").write("")

    with import_path([str(tmpdir)] + sys.path):
        f = filters.ImportableModuleFilter(None)
        skip_parent = f._skip("parent")
        skip_both = f._skip("parent.child")

    # The parent module name is valid because it is not imported, only
    # discovered.
    assert skip_parent is True
    assert "parent" in f.found_modules

    # The child module name is not valid because the parent is
    # imported to find the child and that triggers the side-effect.
    assert skip_both is False
    assert "parent.child" not in f.found_modules


def test_importable_module_with_system_exit(tmpdir):
    path = tmpdir.join("mytestmodule.py")
    path.write('raise SystemExit("exit as side-effect")\n')

    with import_path([str(tmpdir)] + sys.path):
        f = filters.ImportableModuleFilter(None)
        skip = f._skip("mytestmodule")

    # The filter does not actually import the module in this case, so
    # it shows up as a valid word.
    assert skip is True
    assert "mytestmodule" in f.found_modules


def test_pypi_filter_factory():
    f = filters.PyPIFilterFactory()
    assert "sphinxcontrib-spelling" in f.words
    assert "setuptools" in f.words



#!/usr/bin/env python3
#
"""Try to build the Django documentation."""

import argparse
import os
import subprocess
import sys
import tempfile


def doit(*cmd, description="", cwd=None):
    print(f"\n[{description}]\nrunning: {' '.join(cmd)}")
    completed = subprocess.run(cmd, cwd=cwd)
    try:
        completed.check_returncode()
    except subprocess.CalledProcessError as err:
        raise RuntimeError(f"command failed {description}") from err


def try_build(workdir, srcdir, django_repo):
    print(f"working in {workdir}")
    doit(
        "git",
        "clone",
        "--depth",
        "1",
        django_repo,
        "django",
        description="clone django",
        cwd=workdir,
    )
    djangodir = workdir + "/django"
    doit(
        "tox",
        "-e",
        "docs",
        "--notest",
        description="build django docs virtualenv",
        cwd=djangodir,
    )
    doit(
        ".tox/docs/bin/pip",
        "uninstall",
        "-y",
        "sphinxcontrib-spelling",
        description="uninstall packaged sphinxcontrib-spelling",
        cwd=djangodir,
    )
    doit(
        ".tox/docs/bin/pip",
        "install",
        srcdir,
        description="install sphinxcontrib-spelling from source",
        cwd=djangodir,
    )
    doit(
        "tox",
        "-e",
        "docs",
        description="build django docs",
        cwd=djangodir,
    )


def main(args=sys.argv[1:]):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--debug", action="store_true", default=False, help="show full tracebacks"
    )
    parser.add_argument("--src", help="source directory")
    parser.add_argument("--django-repo", default="https://github.com/django/django.git")
    parsed = parser.parse_args(args)

    srcdir = parsed.src
    if not srcdir:
        srcdir = os.path.realpath(os.path.dirname(os.path.dirname(sys.argv[0])))

    try:
        with tempfile.TemporaryDirectory() as dirname:
            try_build(dirname, srcdir, parsed.django_repo)
    except Exception as err:
        if parsed.debug:
            raise
        print(f"ERROR: {err}")
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Tests for SpellingChecker."""

import os

from sphinxcontrib.spelling.checker import SpellingChecker, line_of_index


def test_errors_only():
    checker = SpellingChecker(
        lang="en_US",
        suggest=False,
        word_list_filename=None,
    )
    for word, suggestions, line, offset in checker.check("This txt is wrong"):
        assert not suggestions, "Suggesting"
        assert word == "txt"
        assert line == ""
        assert offset == 0


def test_with_suggestions():
    checker = SpellingChecker(
        lang="en_US",
        suggest=True,
        word_list_filename=None,
    )
    for word, suggestions, line, offset in checker.check("This txt is wrong"):
        assert suggestions, "Not suggesting"
        assert word == "txt"
        assert line == ""
        assert offset == 0


def test_with_wordlist():
    checker = SpellingChecker(
        lang="en_US",
        suggest=False,
        word_list_filename=os.path.join(os.path.dirname(__file__), "test_wordlist.txt"),
    )
    words = [
        word for word, suggestions, line, offset in checker.check("This txt is wrong")
    ]
    assert not words, "Did not use personal word list file"


def test_with_context_line():
    checker = SpellingChecker(
        lang="en_US",
        suggest=False,
        word_list_filename=None,
        context_line=True,
    )

    text = "Line one\nThis txt is wrong\nLine two"
    for word, suggestions, line, offset in checker.check(text):
        assert not suggestions, "Suggesting"
        assert word == "txt"
        assert line == "This txt is wrong"
        assert offset == 1


def test_line_of_index_one_line():
    text = "foo bar baz"
    assert line_of_index(text, 0) == text
    assert line_of_index(text, 5) == text
    assert line_of_index(text, len(text)) == text


def test_line_of_index_multi_line():
    text = "\nfoo\n\nbar baz\n"

    assert line_of_index(text, 0) == ""

    assert line_of_index(text, 1) == "foo"
    assert line_of_index(text, 2) == "foo"
    assert line_of_index(text, 3) == "foo"
    assert line_of_index(text, 4) == "foo"

    assert line_of_index(text, 5) == ""

    assert line_of_index(text, 6) == "bar baz"
    assert line_of_index(text, 12) == "bar baz"
    assert line_of_index(text, 13) == "bar baz"

    assert line_of_index(text, 14) == ""



try:
    # For Python 3.8 and later
    import importlib.metadata as importlib_metadata
except ImportError:
    # For everyone else
    import importlib_metadata

from sphinx.util import logging

from . import asset, builder, directive, domain

logger = logging.getLogger(__name__)


def setup(app):
    version = importlib_metadata.version("sphinxcontrib-spelling")
    logger.info("Initializing Spelling Checker %s", version)
    app.add_builder(builder.SpellingBuilder)
    # Register the 'spelling' directive for setting parameters within
    # a document
    app.add_directive("spelling", directive.LegacySpellingDirective)
    app.add_domain(domain.SpellingDomain)
    # Register an environment collector to merge data gathered by the
    # directive in parallel builds
    app.add_env_collector(asset.SpellingCollector)
    # Report guesses about correct spelling
    app.add_config_value("spelling_show_suggestions", False, "env")
    # Limit the number of suggestions output
    app.add_config_value("spelling_suggestion_limit", 0, "env")
    # Report the whole line that has the error
    app.add_config_value("spelling_show_whole_line", True, "env")
    # Emit misspelling as a sphinx warning instead of info message
    app.add_config_value("spelling_warning", False, "env")
    # Set the language for the text
    app.add_config_value("spelling_lang", "en_US", "env")
    # Set the language for the tokenizer
    app.add_config_value("tokenizer_lang", "en_US", "env")
    # Set a user-provided list of words known to be spelled properly
    app.add_config_value("spelling_word_list_filename", None, "env")
    # Assume anything that looks like a PyPI package name is spelled properly
    app.add_config_value("spelling_ignore_pypi_package_names", False, "env")
    # Assume words that look like wiki page names are spelled properly
    app.add_config_value("spelling_ignore_wiki_words", True, "env")
    # Assume words that are all caps, or all caps with trailing s, are
    # spelled properly
    app.add_config_value("spelling_ignore_acronyms", True, "env")
    # Assume words that are part of __builtins__ are spelled properly
    app.add_config_value("spelling_ignore_python_builtins", True, "env")
    # Assume words that look like the names of importable modules are
    # spelled properly
    app.add_config_value("spelling_ignore_importable_modules", True, "env")
    # Treat contributor names from git history as spelled correctly
    app.add_config_value("spelling_ignore_contributor_names", True, "env")
    # Add any user-defined filter classes
    app.add_config_value("spelling_filters", [], "env")
    # Set a user-provided list of files to ignore
    app.add_config_value("spelling_exclude_patterns", [], "env")
    # Choose whether or not the misspelled output should be displayed
    # in the terminal
    app.add_config_value("spelling_verbose", True, "env")
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
        "version": version,
    }



#
# Copyright (c) 2010 Doug Hellmann.  All rights reserved.
#
"""Spelling checker extension for Sphinx."""

import collections

from docutils.parsers import rst
from sphinx.util import logging

logger = logging.getLogger(__name__)


def add_good_words_to_document(env, docname, good_words):
    # Initialize the per-document good words list
    if not hasattr(env, "spelling_document_words"):
        env.spelling_document_words = collections.defaultdict(list)
    logger.debug("Extending local dictionary for %s with %s", env.docname, good_words)
    env.spelling_document_words[env.docname].extend(good_words)


class SpellingDirective(rst.Directive):
    """Custom directive for passing instructions to the spelling checker.

    .. spelling::

       word1
       word2

    """

    has_content = True

    def run(self):
        env = self.state.document.settings.env

        good_words = []
        for entry in self.content:
            if not entry:
                continue
            good_words.extend(entry.split())
        if good_words:
            add_good_words_to_document(env, env.docname, good_words)

        return []


class LegacySpellingDirective(SpellingDirective):
    def run(self):
        logger.info(
            "direct use of the spelling directive is deprecated, "
            'replace ".. spelling::" with ".. spelling:word-list::"'
        )
        return super().run()



#
# Copyright (c) 2020 Doug Hellmann.  All rights reserved.
#
"""Asset collector for additional spelling terms."""

import collections
import contextlib

from sphinx.environment.collectors import EnvironmentCollector
from sphinx.util import logging

logger = logging.getLogger(__name__)


class SpellingCollector(EnvironmentCollector):
    def clear_doc(self, app, env, docname) -> None:
        with contextlib.suppress(AttributeError, KeyError):
            del env.spelling_document_words[docname]

    def merge_other(self, app, env, docnames, other):
        try:
            other_words = other.spelling_document_words
        except AttributeError:
            other_words = {}

        if not hasattr(env, "spelling_document_words"):
            env.spelling_document_words = collections.defaultdict(list)
        env.spelling_document_words.update(other_words)

    def process_doc(self, app, doctree):
        pass



#!/usr/bin/env python3

import os
import pathlib

import pytest


def require_git_repo(f):
    return pytest.mark.skipif(
        not (pathlib.Path(os.getcwd()) / ".git").is_dir(), reason="Not a git repo"
    )(f)
